/* Program Name            : IO.C                                       */
/*   Original Author       : JLAURET                                    */
/*   Date                  : 07-Oct-1999                                */
/*   Program Description   : Performs IO test using			*/
/*                         : fwrite() and putc()			*/
/* Revision History follows                                             */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "ioperf.h"


/* Other global variables. Does not have to be but saves time.		*/
TypeT		temp;
static clock_t	time1,time2;
static time_t	time3,time4;
static int      blockSZ=0,numIO=0,loopN=0,Slimit=0;
static double	*p;

/* Function cast							*/
static void	erase_file(const char *);
static int	get_size(const char *,int);
static void	init_timer(void);
static void	get_timer(double *,double *);


int main(int argc, char **argv)
{
  int           i,j,k;
  int		test;
  int		limit;
  int		cpid;
  short         Html;
  short		Ptable;
  long int	pos;
  double	maxrand;

  char          flnm[DIM];
  char          label[DIM];
  int		chr;
  FILE          *fo;

  double	cput,realt;
  long int	size[NTESTS];
  double	cpuT[NTESTS],realT[NTESTS],ratioT[NTESTS];

  /* Out in 7s so do not exceed this length 				*/
  const char	*cTests[]={"fwrite","fread","Putc","getc","seek"};

#ifdef FORK_METHOD
  int		child;
#endif

  /* First argument is mandatory and the help is displayed if none	*/
  /* are present ...							*/
  i       = 1;
  Html    = 1==0;
  Ptable  = 1==0;
  label[0]= '\0';
  (void) sprintf(flnm,"ioperf.dat_%d",(int) getpid());


  while( argc > i ) {
    /* Minor argument implementation					*/
    if( strcmp(argv[i],"-h") == 0){
      printf("Syntax is : ioperf [options] where options are \n");
      printf("   -f OutputFileName (current default %s)\n",flnm);
      printf("   -b BlocksPerIO    (default %d)\n",DFLT_BLOCK);
      printf("   -n NumIO          (default %d)\n",DFLT_NUMIO);
      printf("   -l Loops          (default %d)\n",DFLT_NLOOP);
      printf("   -s Nseek          (default %d)\n",DFLT_NSEEK);
      printf("   -m machine_label  (default \"\")\n");
      printf("   -w                print HTML table wrappers\n");
      printf("   -html             output lines in html\n");
      printf("   -h\n\n");
      return SUCCESS;
    } else if ( strcmp(argv[i],"-f") == 0) {
      NEXTARG(i); strcpy(flnm,argv[i]);
    } else if ( strcmp(argv[i],"-b") == 0) {
      NEXTARG(i); blockSZ = atoi(argv[i]);
    } else if ( strcmp(argv[i],"-n") == 0) {
      NEXTARG(i); numIO   = atoi(argv[i]);
    } else if ( strcmp(argv[i],"-l") == 0) {
      NEXTARG(i); loopN	  = atoi(argv[i]);
    } else if ( strcmp(argv[i],"-m") == 0) {
      NEXTARG(i); strcpy(label,argv[i]);
    } else if ( strcmp(argv[i],"-w") == 0) {
      Ptable = (1==1);
    } else if ( strcmp(argv[i],"-html") == 0) {
      Html = 1==1;
    }
    i++;
  }


  /* Some defaults for arguments ...					*/
  if(blockSZ == 0) blockSZ = DFLT_BLOCK;
  if(numIO == 0)   numIO   = DFLT_NUMIO;
  if(loopN == 0)   loopN   = DFLT_NLOOP;
  if(Slimit== 0)   Slimit  = DFLT_NSEEK;



  /* Initialize via pointer on double. Other variables as well.		*/
  p     = (double *) malloc(sizeof(double)*((size_t) blockSZ));
  if(p == NULL){
    fprintf(stderr,"Cannot allocate %d double pointer\n",blockSZ);
    return FAILURE;
  } else {
#ifdef DEBUG    
    fprintf(stderr,"%d arguments read. Allocation OK. ",i-1);
#endif    
    fprintf(stderr,"Scales are  time/%d and clock/%d.\n\n",(int) SCALE,(int) CLK_SCALE);
  }
#ifdef FORK_METHOD
  fprintf(stderr,"Compiled time options included fork() method\n");
#else
  fprintf(stderr,"Direct IO method is used.\n");
#endif

  for(i=0 ; i < NTESTS; i++){
    cpuT[i]  = 0.0;
    realT[i] = 0.0;
    ratioT[i]= 0.0;
    size[i]  = 0;
  }

  limit = blockSZ*((int) sizeof(p));
  Slimit= MIN(Slimit,limit);


  /* rand() and random() returns a number between 0 and 		*/
  /* (2**31-1)=	INT_MAX							*/
  maxrand = (double) INT_MAX;


  /* Get ready to start. Last info.					*/
  fprintf(stderr,"Time Results in seconds, Rate in %s, real then cpu-based\n",BYTEST);
  fprintf(stderr,"Blocks per IO %d, Sizeof double is %d, %d IOs, %d loop%s\n",
    blockSZ,(int) sizeof(double),numIO,loopN,loopN==1?"":"s");
  fprintf(stderr,"%d/%d seek will be done\n",Slimit,limit);
  fprintf(stderr,"Output file will be %s\n\n",flnm);

  fprintf(stderr,"IO tester begins\n");

  for(j = 0 ; j < loopN ; j++){
    test = 0;
    /* ---------------------------------------------------------------- */
    /* Unformatted IO. WRITE first.					*/
    init_timer();
    BEGIN_TEST
      if( (fo = fopen(flnm,"w")) != NULL){
	for(i = 0 ; i < numIO ; i++){
	  for( k = 0 ; k < blockSZ ; k++){
	    p[k] = (double) (i*blockSZ+k);
	  }
	  (void) fwrite(p,sizeof(double),(size_t) blockSZ,fo);
	}
	(void) fsync(fileno(fo));
	(void) fclose(fo);
      } else {
	fprintf(stderr," Could not open %s\n",flnm);
	RETURN(FAILURE);
      }
    END_TEST;

    get_timer(&realt,&cput);
    if( realt != 0.0 /*~ExactCmp*/){
      realT[test] += realt;
      cpuT[test]  += cput;
      ratioT[test]+= cput/realt*PRCT;
    }


    /* Do test 2. READ (file exists now)				*/
    test++;
    init_timer();
    BEGIN_TEST
      if( (fo = fopen(flnm,"r")) != NULL){
	for(i=0 ; i < numIO ; i++){
	  (void) fread(p, sizeof(double), (size_t) blockSZ, fo);
#ifdef DEBUG
	  /* This DEBUG block was added because I got a rate much greater
	     than what is actually possible. I thought something was wrong
	     but the recovered data prints OK.				*/
	  for( k = 0 ; k < blockSZ ; k++){
	    printf("%f ",p[k]);
	  }
	  printf("\n");
#endif
	}
	(void) fclose(fo);
      } else {
	fprintf(stderr," Could not reopen %s.\n",flnm);
	RETURN(FAILURE);
      }
    END_TEST;

    get_timer(&realt,&cput);
    if( realt != 0.0 /*~ExactCmp*/){
      realT[test] += realt;
      cpuT[test]  += cput;
      ratioT[test]+= cput/realt*PRCT;
    }

    /* Check file size for report. Do that only once			*/
    if(j == 0){
      size[test]  = get_size(flnm,test);
      size[test-1]= size[test];
    }
    erase_file(flnm);



    /* ---------------------------------------------------------------- */
    /* Test is now the worst case of formatted IO			*/
    test++;
    init_timer();
    BEGIN_TEST
      if( (fo = fopen(flnm,"w")) != NULL){
	for(i = 0 ; i < numIO ; i++){
	  for( k = blockSZ*((int) sizeof(double)) ; k > 0 ; k--){
	    (void) fputc(k & 0x7f ,fo);
	  }
	}
	(void) fsync(fileno(fo));
	(void) fclose(fo);
      } else {
	fprintf(stderr," Could not open %s\n",flnm);
	RETURN(FAILURE);
      }
    END_TEST;


    get_timer(&realt,&cput);
    if( realt != 0.0 /*~ExactCmp*/){
      realT[test] += realt;
      cpuT[test]  += cput;
      ratioT[test]+= cput/realt*PRCT;
    }


    test++;
    init_timer();
    BEGIN_TEST
      if( (fo = fopen(flnm,"r")) != NULL){
	for(i = 0 ; i < numIO ; i++){
	  for( k = blockSZ*((int) sizeof(double)) ; k > 0 ; k--){
	    chr = fgetc(fo);
	  }
	}
	(void) fclose(fo);
      } else {
	fprintf(stderr," Could not reopen %s.\n",flnm);
	RETURN(FAILURE);
      }
    END_TEST;

    get_timer(&realt,&cput);
    if( realt != 0.0 /*~ExactCmp*/){
      realT[test] += realt;
      cpuT[test]  += cput;
      ratioT[test]+= cput/realt*PRCT;
    }
    if(j == 0){
      size[test]  = get_size(flnm,test);
      size[test-1]= size[test];
    }


    /* Next test is a random seek in an existing file so we do	*/
    /* NOT delete the file after the preceding test.		*/
    test++;
    size[test]= size[test-1]/limit*Slimit;
    init_timer();
    BEGIN_TEST
      if( (fo = fopen(flnm,"r")) != NULL){
	for(i = 0 ; i < numIO ; i++){
	  /* We will not cover the entire file but at least do	*/
	  /* a good sampling of the number of bytes it does 	*/
	  /* contain ...					*/
	  for( k = Slimit ; k > 0 ; k--){
	    pos = rand();
	    (void) fseek(fo,0,(int) ((pos+1)*limit/maxrand));
	  }
	}
	(void) fclose(fo);
      } else {
	fprintf(stderr," Could not reopen %s.\n",flnm);
	RETURN(FAILURE);
      }
    END_TEST;

    get_timer(&realt,&cput);
    if( realt != 0.0 /*~ExactCmp*/){
      realT[test] += realt;
      cpuT[test]  += cput;
      ratioT[test]+= cput/realt*PRCT;
    }

    erase_file(flnm);



    /* Just to force us to check that out				*/
    if(test > NTESTS){
      fprintf(stderr," Inconsistent number of tests\n");
      return FAILURE;
    } else {
      k = j+1;
      fprintf(stderr," Pass %5d.",k);
      if(k==1)	fprintf(stderr," File sizes 1=%d 3=%d\n",size[0],size[2]);
      else	fprintf(stderr,"\n");
    }

    /* Else, display stat						*/
    if(Html){
      if(j == 0 && Ptable){
#include "table_head.c"
      }
      printf("<TR><TD ALIGN=\"right\">%s</TD><TD ALIGN=\"right\">%d</TD>\n",
	label,(int) (0.5+ceil(size[0]/BYTESCL)));
    }
    for(i=0 ; i < NTESTS ; i++){
      if( cpuT[i] == 0.0  /*~ExactCmp*/ ||
	  realT[i] == 0.0 /*~ExactCmp*/  ){
	fprintf(stderr," %7s :: Realtime or CPUtime is too small.\n",
	  cTests[i]);
      } else {
	if(Html){
	  printf("<TD ALIGN=\"right\">%14.2f</TD><TD ALIGN=\"right\">%14.2f</TD><TD ALIGN=\"right\">%6.2f</TD>\n",
	    (k*size[i]/realT[i])/BYTESCL,(k*size[i]/cpuT[i])/BYTESCL,ratioT[i]/k);
	} else {
	  printf(" %7s :: Times=%7.2f %7.2f (%6.2f%%) Rates=%14.2f %14.2f\n",
	    cTests[i],realT[i]/k,cpuT[i]/k,ratioT[i]/k,
	    (k*size[i]/realT[i])/BYTESCL,(k*size[i]/cpuT[i])/BYTESCL);
	}
      }
    }
    if(Html){
      printf("</TR>\n");
    }
  }
  if(Html && Ptable){
#include "table_foot.c"
  }
  return SUCCESS;
}


/* -------------------------------------------------------------------  */
/* File manipulation.							*/
/* -------------------------------------------------------------------  */

static void erase_file(const char *flnm)
{
  FILE *ft;

  /* Above supports version number (i.e. VMS file system)		*/
  /* Note that deleting files in between operations makes this test	*/
  /* a bit more system indepedant (remember advFs+raid combo)		*/
  while( (ft = fopen(flnm,"r")) != NULL ){
    (void) fclose(ft);
    (void) UNLINK(flnm);
  }
}


static int get_size(const char *flnm, int tst)
{
  int		size;
  struct stat	info;

  /* note that the calculated size may be different from the real size	*/
  /* depending on the file system. This is not unique to VMS fs since	*/
  /* advFs behaves the same way.					*/
  (void) stat(flnm,&info);
  if( info.st_size != 0){
    size = (int) info.st_size;
  } else {
    size = (int) sizeof(double)*blockSZ*numIO;
    fprintf(stderr," stat() failed. Taking aproximate size\n");
  }
  return size;
}

/* -------------------------------------------------------------------  */
/* Time related functions.						*/
/* -------------------------------------------------------------------  */

static void init_timer()
{
  /* sleep is to allow buffering flush */
  time3       = TIMES(&temp);
  time1       = clock();
  if(time1 == -1){
    fprintf(stderr,"clock() returned -1\n");
    abort();
  }
}

static void get_timer(double *realt, double *cput)
{
  time2 = clock();
  time4 = TIMES(&temp);

  *realt= ((double) (time4-time3))/SCALE;
  *cput	= ((double) (time2-time1))/CLK_SCALE;

}

/* ----------------------- :                                            */
/*   Date of Revision      :  8-OCT-1999                                */
/*   Change Author         : JLAURET                                    */
/*   Purpose of Revision   : Tested on Linux/OpenVMS and DEC OSF1	*/
/*                         :                                            */

/* ----------------------- :                                            */
/*   Date of Revision      : 22-JUN-2000                                */
/*   Change Author         : JLAURET                                    */
/*   Purpose of Revision   : Added -m and -html as in Bonnie    	*/
/*                         :                                            */
